#include <errno.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <pty.h>
#include <sys/fcntl.h>
#include "proc.h"

typedef struct pr
{
    struct pr *prev,*next;
    pid_t pid;
    int status;
    int ecode; /* exit reason (exit/signal) */
    int earg;  /* exit code or signal name */
    int opts;

 //      For          For
 //     Child        Parent  
 //
    int STDIN_r,     STDIN_w;
    int STDOUT_w,    STDOUT_r;
    int STDERR_w,    STDERR_r;

    int method;  // 0= pipe  1= pty
    int ptyfd;
} pr_t;

static pr_t head= { &head, &head };
static int sig_installed=0;
static struct sigaction  old_sa;



int proc_kill(int id,int sig)
{
  if (kill(id,sig))
     return -errno;
  return 0;
}

/***********************************************************************
 There are two methods to choose from.
   - pipe method
   - pty method
 For the pipe method, all necessary fds are mapped to pipes and the
 unused ones are mapped to /dev/null.

 For the pty method, the input and output fds are automatically mapped
 to the pseudo-terminal, we don't do anything special for that.

 The error stream is mapped to /dev/null if not required. Otherwise,
 it's mapped to a pipe.

 This way, the error stream never intermingles with the output stream.
 ***********************************************************************/

static int map_fd(int d_old, int d_new)
{
  if (d_old<0) return 1;
  while(dup2(d_old,d_new)<0) 
    if (errno!=EINTR) return 1; 
  return 0;
}

static int close_fd(int *FD)
{
  if (!FD) return 0;
  if (*FD<0) return 0;
  while(close(*FD)) { if (errno==EINTR) continue; *FD= -1; return 1; }
  *FD= -1;
  return 0;
}

static int map_fd_and_close(int fdold, int fdnew, int *cl1, int *cl2)
{
  int R;
  R= 0;
  R |= map_fd(fdold,fdnew);
  R |= close_fd(cl1);
  R |= close_fd(cl2);
  return R;
}


/***********************************************************************
 The following functions are called by the child process. They arrange
 the mappings from standard in/out/err streams.
 ***********************************************************************/

static int child_map_fds_pipe(pr_t *p)
{
  int devnull;
  int R;

  R= 0;
  devnull= -1;
  if (!(p->opts&PROC_ERR) || !(p->opts&PROC_OUT) || !(p->opts&PROC_INP))
    devnull= open("/dev/null", O_RDWR);

  if (p->opts & PROC_INP)
    R|= map_fd_and_close(p->STDIN_r, 0 , &p->STDIN_w, &p->STDIN_r);
  else
    R|= map_fd(devnull,0);

  if (p->opts & PROC_OUT)
    R|= map_fd_and_close(p->STDOUT_w, 1, &p->STDOUT_w, &p->STDOUT_r);
  else
    R|= map_fd(devnull,1);

  if (p->opts & PROC_ERR)
    R|= map_fd_and_close(p->STDERR_w, 2, &p->STDERR_w, &p->STDERR_r);
  else
    R|= map_fd(devnull,2);

  close_fd(&devnull);
  return R;
}

static int child_map_fds_pty(pr_t *p)
{
  int devnull,R;

  if (p->opts & PROC_ERR)
    R= map_fd_and_close(p->STDERR_w, 2, &p->STDERR_w, &p->STDERR_r);
  else
  {
    devnull= open("/dev/null", O_RDWR);
    R= map_fd(devnull,2);
    close_fd(&devnull);
  }
  return R;
}

static int child_map_fds(pr_t *p)
{
  if (p->method) return child_map_fds_pty(p);
            else return child_map_fds_pipe(p);
}


/***********************************************************************
 Similarly, when the child is created, the parent process also has to
 close some file descriptors. This is easier, as there is no dup2 stuff.
 ***********************************************************************/

static int parent_close_unused_fds(pr_t *p)
{
  int R;
  R= 0;
  if (!p->method)
  {
    if (p->opts&PROC_INP) R|= close_fd(&p->STDIN_r);
    if (p->opts&PROC_OUT) R|= close_fd(&p->STDOUT_w);
  }
  if (p->opts&PROC_ERR) R|= close_fd(&p->STDERR_w);
  return R;
}


/***********************************************************************
 When a process is finished, we must close the corresponding descriptors
 in the parent. This is done automatically for the child since it already
 exits. In fact, we can close everything since we're done with them.
 ***********************************************************************/
static int close_all_fds(pr_t *p)
{
  int R;
  R= 0;
  R|= close_fd(&p->STDIN_w);
  R|= close_fd(&p->STDIN_r);
  R|= close_fd(&p->STDOUT_r);
  R|= close_fd(&p->STDOUT_w);
  R|= close_fd(&p->STDERR_r);
  R|= close_fd(&p->STDERR_w);
  R|= close_fd(&p->ptyfd);
  return R;
}


/***********************************************************************
 We make the pipes in different ways for the two methods as described above.
 For the pty case, the input and output fds are made by the forkpty
 function. Only the error stream needs to be redirected.
 ***********************************************************************/

static int make_pipe(int *rfd, int *wfd)
{
  int fds[2];
  if (pipe(fds)) return 1;
  *rfd= fds[0];
  *wfd= fds[1];
  return 0;
}

static int make_fds(pr_t *p)
{
  int R;
  R= 0;
  if (!p->method)
  {
    if (p->opts&PROC_INP) R|= make_pipe(&p->STDIN_r, &p->STDIN_w);
    if (p->opts&PROC_OUT) R|= make_pipe(&p->STDOUT_r, &p->STDOUT_w);
  }
  if (p->opts&PROC_ERR) R|= make_pipe(&p->STDERR_r, &p->STDERR_w);
  return R;
}


static void child(pr_t *p,char *fn,char **argv,int opts) 
{
  if (child_map_fds(p)) exit(127);
  execvp(fn,argv);
  exit(126);
}

static pr_t *find(int pid)
{
  pr_t *p;
  for(p=head.next;p!=&head;p=p->next) if (p->pid==pid) return p;
  return NULL;
}


static void child_sig(int sno,siginfo_t *si,void *dummy)
{
  int pid= si->si_pid;
  int status= si->si_status;
  int code= si->si_code;
  pr_t *p;
  p= find(pid);
  if (!p)
    return ;
  switch(code) {
    case CLD_EXITED:
    case CLD_KILLED:
    case CLD_DUMPED:
      p->status= PROC_DEAD;
      p->ecode= code;
      p->earg= status;
      break;
    case CLD_TRAPPED:
    case CLD_STOPPED:
      p->status= PROC_STOPPED;
      break;
    case CLD_CONTINUED:
      p->status= PROC_RUNNING;
      break;
  }
}

static void install_handler() {
  struct sigaction sa;
  if (sig_installed)
    return ;
  sa.sa_sigaction= child_sig;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags= SA_SIGINFO;
  sa.sa_restorer= 0;
  while(sigaction(SIGCHLD, &sa, &old_sa) && errno==EINTR) ;
  sig_installed= 1;
}


static void remove_node(pr_t *p)
{
  p->prev->next= p->next;
  p->next->prev= p->prev;
  free(p);
}


static pr_t *make_node()
{
  pr_t *p;

  p= calloc(1,sizeof(*p));
  if (!p) return NULL;
  p->prev= head.prev;
  p->next= &head;
  head.prev->next= p;
  head.prev= p;

  p->STDIN_w= p->STDIN_r= p->STDOUT_w= p->STDOUT_r
            = p->STDERR_w= p->STDERR_r= p->ptyfd = -1;
  return p;
}

static int fork_child(pr_t *p)
{
  struct winsize win = { .ws_col = 80, .ws_row = 24,
                         .ws_xpixel = 480, .ws_ypixel = 192, };
  if (p->method)
    p->pid= forkpty(&p->ptyfd, NULL, NULL, &win);
  else
    p->pid= fork();
  return (p->pid<0);
}

int proc_start(char *fn,char **argv,int opts)
{
  int eno;
  pr_t *p;
  
  install_handler();
  
  if (!(p=make_node())) return -ENOMEM;
  p->status= PROC_RUNNING;

  if (opts & PROC_PTY) p->method= 1; 
  p->opts= opts;

  if (make_fds(p)) goto sockerr; 

  if (fork_child(p)) goto sockerr;
  if (p->pid==0) child(p,fn,argv,opts);
  parent_close_unused_fds(p);
  return p->pid;

sockerr:
  eno= errno;
  close_all_fds(p);
  remove_node(p);
  return -eno;
}

static void interp_status(int s,pr_t *p)
{
  if (WIFEXITED(s))
  {
    p->ecode= CLD_EXITED;
    p->earg= WEXITSTATUS(s);
  } 
  else if (WIFSIGNALED(s))
  {
    p->ecode= CLD_DUMPED;
    p->earg= WTERMSIG(s);
  }
  p->status= PROC_DEAD;
}

int proc_end(int id,int *st) {
  int status;
  pr_t *p;
  int c,a;
  p= find(id);
  if (!p)
    return 0;
  close_all_fds(p);
  while(waitpid(id,&status,0)<0 && errno==EINTR)
    ;
  if (p->status!=PROC_DEAD)
    interp_status(status,p);
  c= p->ecode;
  a= p->earg;
  remove_node(p);
  if (c==CLD_KILLED || c==CLD_DUMPED || (c==CLD_EXITED && a!=0)) 
  {
    if (st) *st= a;
    return c;
  }
  return 0;
}




int proc_input(int id)
{
  pr_t *p;
  return (p=find(id)) ? 
             (p->method ? p->ptyfd : p->STDIN_w)
            : -ESRCH;
}

int proc_output(int id)
{
  pr_t *p;
  return (p=find(id)) ? 
             (p->method ? p->ptyfd : p->STDOUT_r)
            : -ESRCH;
}

int proc_error(int id)
{
  pr_t *p;
  return (p=find(id)) ? p->STDERR_r : -ESRCH ;
}

int proc_islive(int id) 
{
  pr_t *p= find(id);
  return p ? (p->status!=PROC_DEAD) : 0;
}

typedef struct buffer {
  struct buffer *next;
  char data[4000];
  int len;
} buffer_t;


static buffer_t *do_read(int fd, fd_set *set, buffer_t *tail,int *ende)
{
  int r;
  *ende= 0;
  if (fd!=-1 && FD_ISSET(fd,set)) {
    int l= sizeof(tail->data)- tail->len;
    while(1) {
      r= read(fd, tail->data+tail->len, l);
      if (r<0 && errno==EINTR)
        continue;
      if (r<=0) {
        *ende= 1;
        return tail; /* FIXME: think of something */
      }
      break;
    }
    tail->len+= r;
    if (tail->len==sizeof(tail->data)) {
      buffer_t *b= malloc(sizeof(*b));
      b->len= 0;
      b->next= 0;
      tail->next= b;
      tail= b;
    }
  }
  return tail;
}

static void collect_buffer(char **d,size_t *l, buffer_t *h)
{
  size_t s= 0;
  char *D;
  buffer_t *p,*n;

  if (!d) return ;

  for(p=h;p;p=p->next) s+= p->len;
  D= malloc(s+1);
  *d= D;

  for(p=h;p;p=p->next)
  {
    memcpy(D, p->data, p->len);
    D+= p->len;
  }
  D[0]= 0;

  for(p=h->next;p;p=n)
  {
    n= p->next;
    free(p);
  }
  *l= s;
}

static void add_set(int *max, fd_set *set,int fd) {
  FD_SET(fd,set);
  if (fd>*max) *max= fd;
}


/* one-shot */
int proc_run(char *fn,char **argv,int *arg,
             char *inp,size_t inp_l,
             char **outp,size_t *outp_l,
             char **err,size_t *err_l)
{
  int input=0;
  int inp_fd=-1,outp_fd=-1,err_fd=-1;
  int r;
  int opts=0;
  int ch;
  buffer_t errhead;
  buffer_t outhead;
  buffer_t *errb;
  buffer_t *outb;
  pr_t *p;
  if (inp) opts |= PROC_INP;
  if (outp) { *outp= NULL; opts |= PROC_OUT; }
  if (err) { *err= NULL; opts |= PROC_ERR; }
  ch= proc_start(fn,argv, opts);
  if (ch<0)
    return ch;
  if (inp) inp_fd= proc_input(ch);
  if (outp) outp_fd= proc_output(ch);
  if (err) err_fd= proc_error(ch);
  p= find(ch);
  if (!p)
    exit(fprintf(stderr,"internal error in proc library\n"));
  errhead.next= 0;
  outhead.next= 0;
  errhead.len= 0;
  outhead.len= 0;
  errb= &errhead;
  outb= &outhead;
  if (inp_fd!=-1)
    input= 1;
  while(p->status!=PROC_DEAD && (inp_fd!=-1 || err_fd!=-1 || outp_fd!=-1))
  {
    int e;
    int status;
    fd_set read_fd,write_fd;
    int max= 0;

    FD_ZERO(&write_fd);
    FD_ZERO(&read_fd);
    if (input)       add_set(&max, &write_fd, inp_fd);
    if (err_fd!=-1)  add_set(&max, &read_fd, err_fd);
    if (outp_fd!=-1) add_set(&max, &read_fd, outp_fd);

    if (waitpid(ch,&status,WNOHANG)==ch)
    {
      interp_status(status,p);
      break;
    }
    r= select( max+1,
               (err_fd!=-1 || outp_fd!=-1) ? &read_fd : NULL,
               input ? &write_fd : NULL, NULL,
               NULL);

    if (r<0) {
      if (errno!=EINTR)
        exit(fprintf(stderr,"select failed: %s\n",strerror(errno)));
      continue;
    }
    if (r==0) {
      if (p->status==PROC_DEAD)
        break;
      continue;
    }
    outb= do_read(outp_fd, &read_fd, outb,&e);
    if (e) outp_fd= -1;
    errb= do_read(err_fd, &read_fd, errb,&e);
    if (e) err_fd= -1;
    if (input && FD_ISSET(inp_fd, &write_fd))
    {
      int r;
      int l= 1024;
      if (l>inp_l) l= inp_l;
      while( (r=write(inp_fd,inp,l))<0 && errno==EINTR ) ;
      if (r<0) break;
      inp_l-= r;
      inp+=r;
      if (inp_l==0)
      {
        input= 0;
        inp_fd= -1;
      }
    }
  }

  /* read the remaining data */
  while(err_fd!=-1 || outp_fd!=-1) 
  {
    int e;
    fd_set read_fd;
    int max= 0;
    struct timeval tm;
    FD_ZERO(&read_fd);
    if (err_fd!=-1)  add_set(&max, &read_fd, err_fd);
    if (outp_fd!=-1) add_set(&max, &read_fd, outp_fd);
    tm.tv_sec= 0;
    tm.tv_usec= 0;

    r= select( max+1,  &read_fd ,  NULL, NULL, &tm);

    if (r<0) 
    {
      if (errno!=EINTR)
        exit(fprintf(stderr,"select failed: %s\n",strerror(errno)));
      continue;
    }
    if (r==0) break;
    outb= do_read(outp_fd, &read_fd, outb,&e);
    if (e) outp_fd= -1;
    errb= do_read(err_fd, &read_fd, errb,&e);
    if (e) err_fd= -1;
  }

  if (outp) collect_buffer(outp,outp_l, &outhead);
  if (errb) collect_buffer(err,err_l, &errhead);
  return proc_end(ch,arg);
}

